/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.syncclient.spds;

import java.io.IOException;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import sync4j.syncclient.common.Base64;
import sync4j.syncclient.common.StringTools;

import sync4j.syncclient.sps.DataAccessException;
import sync4j.syncclient.sps.DataStore;
import sync4j.syncclient.sps.Record;
import sync4j.syncclient.util.PagedVector;
import sync4j.syncclient.util.StaticDataHelper;

/**
 * The <i>SyncManager</i> is the contact point between a host application and the
 * synchronization engine. It is designed to hidden as much as possible the
 * details of the synchronization logic, protocol, communication and so on.
 *
 * @author Fabio Maggi @ Funambol
 * @version  $Id: SyncManagerImpl.java,v 1.23 2006/02/07 09:47:17 luigiafassina Exp $
 */
public class SyncManagerImpl {

    //---------------------------------------------------------------- Constants

    public final static int
        ALERT_CODE_ONE_WAY_FROM_SERVER               = 204             ;
    public static final int
        ALERT_CODE_REFRESH_FROM_SERVER               = 205             ;
    public static final int
        ALERT_CODE_SLOW                              = 201             ;
    public static final int
        STATUS_CODE_FORBIDDEN                        = 403             ;
    public static final int
        STATUS_CODE_OK                               = 200             ;
    public static final int
        STATUS_CODE_AUTHENTICATION_ACCEPTED          = 212             ;
    public static final int
        STATUS_REFRESH_REQUIRED                      = 508             ;

    private static final String XML_ALERT            = "/xml/alert.xml";
    private static final String XML_INITIALIZATION   = "/xml/init.xml" ;
    private static final String XML_MODIFICATIONS    = "/xml/mod.xml"  ;
    private static final String XML_MAPPING          = "/xml/map.xml"  ;

    private static final char   RECORD_STATE_NEW     = 'N'             ;
    private static final char   RECORD_STATE_DELETED = 'D'             ;
    private static final char   RECORD_STATE_UPDATED = 'U'             ;

    private static final String TAG_ALERT            = "Alert"         ;
    private static final String TAG_ADD              = "Add"           ;
    private static final String TAG_CMD              = "Cmd"           ;
    private static final String TAG_DATA             = "Data"          ;
    private static final String TAG_DELETE           = "Delete"        ;
    private static final String TAG_ITEM             = "Item"          ;
    private static final String TAG_LOC_URI          = "LocURI"        ;
    private static final String TAG_REPLACE          = "Replace"       ;
    private static final String TAG_STATUS           = "Status"        ;
    private static final String TAG_SYNC             = "Sync"          ;
    private static final String TAG_SYNCBODY         = "SyncBody"      ;
    private static final String TAG_SYNCHDR          = "SyncHdr"       ;
    private static final String TAG_SYNCML           = "SyncML"        ;
    private static final String TAG_TARGET           = "Target"        ;
    private static final String MAX_MSG_SIZE         = "40768"         ;

    //---------------------------------------------------------------- Private data

    private Hashtable   serverAlerts          = null ;

    private String      alertXML              = null ;
    private String      clientInitXML         = null ;
    private String      modificationsXML      = null ;
    private String      mappingXML            = null ;
    
    private String      modificationsXMLTemplate = null;

    private String      serverUrl             = null ;
    private String      serverUrlStatus       = null ;
    private String      gatewayApn            = null ;
    private String      gatewayIp             = null ;
    private String      gatewayPort           = null ;
    private String      gatewayUrl            = null ;

    private String      login                 = null ;

    private String      sourceUri             = null ;
    private String      sourceType            = null ;
    private String      sourceName            = null ;
    private String      sessionID             = null ;
    private Hashtable   mappings              = null ;

    private DataStore   dataStore             = null ;

    private long        lastTimestamp                ;
    private long        nextTimestamp                ;

    private boolean     moreRecordsDelete     = true ; 
    private boolean     moreRecordsNew        = true ; 
    private boolean     moreRecordsUpdate     = true ; 
    
    Runtime rt = null;

    //------------------------------------------------------------- Constructors

    /**
     * Creates a SyncManagerImpl.
     * The datastore is identified by the dataStoreName parameter
     *
     * @param dataStoreName The dataStore to Sync
     *
     */
    protected SyncManagerImpl (String dataStoreName ,
                               String sourceType    ,
                               String sourceName) {

        this.sourceUri      = dataStoreName                          ;
        this.sourceType     = sourceType                             ;
        this.sourceName     = sourceName                             ;
        this.mappings       = new Hashtable()                        ;
        this.dataStore      = DataStore.getDataStore(dataStoreName)  ;


    }

    //----------------------------------------------------------- Public methods

    public void setServerUrl (String serverUrl)      {
        this.serverUrl = serverUrl;
    }

    public void setLogin (String login)              {
        this.login = login;
    }

    public void setGatewayApn (String gatewayApn)    {
        this.gatewayApn = gatewayApn;
    }

    public void setGatewayIp (String gatewayIp)      {
        this.gatewayIp = gatewayIp;
    }

    public void setGatewayPort (String gatewayPort)  {
        this.gatewayPort = gatewayPort;
    }

    /** Synchronize synchronization sources.
     * @throws SyncException if an error occurs during synchronization
     * @throws AuthenticationException if the server responded with "non authorized" return code
     */
    public void sync() throws SyncException, AuthenticationException, UpdateException {

        String response = "";

        if (gatewayApn != null && gatewayApn.length() > 0) {
            gatewayUrl = ";WAPGatewayIP="   +
                         gatewayIp          +
                         ";WAPGatewayAPN="  +
                         gatewayApn         +
                         ";WAPGatewayPort=" +
                         gatewayPort        ;
        } else {
            gatewayUrl = "";
        }

        serverUrl = serverUrl + gatewayUrl;

        try {
            lastTimestamp = dataStore.getLastTimestamp();
        } catch (DataAccessException e) {
            throw new SyncException(e.getMessage());
        }

        nextTimestamp = System.currentTimeMillis();

        try {
            loadResources();
        } catch (IOException e) {
            throw new UpdateException( "Error: ("
            + e.getClass().getName()
            + "): "
            + e.getMessage()
            );
        }

        sessionID = String.valueOf(System.currentTimeMillis());

        prepareInizializationMessage();

        StaticDataHelper.log("\n\n\n");
        StaticDataHelper.log("sending:\n" + clientInitXML);

        response = syncInitialization();

        checkServerAlerts(response);

        // start debug
        StaticDataHelper.log("\n\n\n");
        Enumeration keys = serverAlerts.keys();
        while (keys.hasMoreElements()) {
            String keyStr = (String) keys.nextElement();
            StaticDataHelper.log("source - Name: " + keyStr + "| Status: " + serverAlerts.get(keyStr));
        }
        // end debug

        StaticDataHelper.log("\n\n\n");
        StaticDataHelper.log("receiving\n:" + response);

        serverUrlStatus = serverUrl;
        serverUrl       = response.substring(response.indexOf("<RespURI>") + 9,
                                             response.indexOf("</RespURI>")) + gatewayUrl;

        //
        // execute init dataStore operation
        //
        dataStore.startDSOperations();

        //
        // Notifies the sources that the synchronization is going to begin
        //

        boolean sendingClientModifications = true ;
        boolean done                       = false;
        do {

            int alertCode = -1;

            response = null;
            sendingClientModifications =
                prepareModificationMessage(sendingClientModifications);

            StaticDataHelper.log("\n\n\n\n\n\n\n");
            StaticDataHelper.log("sending\n:" + modificationsXML);
  
            response = syncModifications();
            
            StaticDataHelper.log("\n\n\n\n\n\n\n");
            StaticDataHelper.log("receiving\n:" + response);

            processModifications(response);
           
            done = 
                ((response.indexOf("<Final/>") >= 0) 
                || (response.indexOf("</Final>") >= 0));
        } while (!done);

        //
        // We still have to send the latest mappings
        //
        prepareMappingMessage();

        StaticDataHelper.log("\n\n\n\n\n\n\n");
        StaticDataHelper.log("sending\n:" + mappingXML);

        response = syncMapping();
    
        StaticDataHelper.log("\n\n\n\n\n\n\n");
        StaticDataHelper.log("receiving\n:" + response);

        //
        // Set the last anchor to the next timestamp for all the sources that
        // has been synchronized
        //

        try {
            dataStore.commitDSOperations();
            dataStore.setLastTimestamp(nextTimestamp);
        } catch (DataAccessException e) {
            throw new SyncException(e.getMessage());
        }

    }

    //---------------------------------------------------------- Private methods

    /**
     * Sync initialization.
     *
     * @throws SyncException, AuthenticationException, UpdateException
     * @return sync initialization response
     **/
    private String syncInitialization()
    throws SyncException, AuthenticationException, UpdateException {

        String response = "";

        StaticDataHelper.log("Sync Initialization...");

        try {
            response = postRequest(clientInitXML);
        } catch(SyncException e) {
            String msg = "Error Sync Initialization: " + e.getMessage();
            StaticDataHelper.log(msg);
            throw e;
        } catch (Exception e) {
            String msg = "Error Sync Initialization: " + e.getMessage();
            StaticDataHelper.log(msg);
            throw new SyncException(msg);
        }

        checkStatus(response, TAG_SYNCHDR );
        checkStatus(response, TAG_ALERT   );

        StaticDataHelper.log("Initialization done!");

        return response;
    }


    /**
     * Synch Modification.
     *
     * @throws UpdateException
     * @return sync modification response
     **/
    private String syncModifications()
    throws SyncException {

        String response = null;

        StaticDataHelper.log("Sync Modifications...");

        try {
            response = postRequest(modificationsXML);
        } catch(SyncException e) {
            String msg = "Error Sync Modification: " + e.getMessage();
            throw new SyncException(msg);
        } catch (Exception e) {
            String msg = "Error Sync Modification: " + e.getMessage();
            StaticDataHelper.log(msg);
            throw new SyncException(msg);
        }
        StaticDataHelper.log("Modification done!");
        return response;
    }

    /**
     * Send alert message.
     *
     * @throws UpdateException
     * @return sync modification response
     **/
    private String sendAlertMessage()
    throws SyncException {

        String response = null;

        StaticDataHelper.log("Sync Modifications...");

        try {
            response = postRequest(alertXML);
        } catch(SyncException e) {
            String msg = "Error Modifications: " + e.getMessage();
            throw new SyncException(msg);
        } catch (Exception e) {
            String msg = "Error Modifications: " + e.getMessage();
            StaticDataHelper.log(msg);
            throw new SyncException(msg);
        }
        StaticDataHelper.log("Modification done!");

        return response;
    }


    /**
     * Sync LUID-GUID mapping.
     *
     * @throws UpdateException
     * @return sync modification response
     **/
    private String syncMapping()
    throws SyncException {
        String response = null;
        
        StaticDataHelper.log("Sending mapping...");

        try {
            return postRequest(mappingXML);
        } catch (Exception e) {
            String msg = "Error Sync Mapping: " + e.getMessage();
            StaticDataHelper.log(msg);
            throw new SyncException(msg);
        }
    }

    /**
     * Posts the given message to the url specified by <i>url</i>.
     *
     * @param request the request msg
     * @return the url content as a byte array
     * @throws SyncException in case of network errors
     **/
    private String postRequest(String request)
    throws SyncException {

        try {

            SyncMLClientImpl syncMLClient = new SyncMLClientImpl(serverUrl);

            return syncMLClient.sendMessage(request);

        } catch (NetworkException e) {
            String msg = "Synchronization failed: no network available";
            StaticDataHelper.log(msg);
            throw new SyncException(msg);
        } catch (IOException e) {
            String msg = "Synchronization failed, connection broken";
            StaticDataHelper.log(msg);
            throw new SyncException(msg);
        } catch (Exception e) {
            String msg = "Error sending the request: network error";
            StaticDataHelper.log(msg);
            throw new SyncException(msg);
        } 
    }

    /**
     * Checks if the given response message is authenticated from the server.
     *
     * @param msg the message to be checked
     * @param statusOf the command of which we want to check the status
     *
     * @throws AuthenticationException in case the request has not been
     *                                 authenticated by the server
     * @throws SyncException in case of other errors
     **/
    private void checkStatus(String msg, String statusOf)
    throws SyncException, AuthenticationException, UpdateException {

        Vector xmlMsg = new Vector();

        xmlMsg.addElement(msg);

        Vector statusTag = getXMLTag(getXMLTag(getXMLTag(xmlMsg, TAG_SYNCML),TAG_SYNCBODY), TAG_STATUS);

        int l = statusTag.size();

        for (int i=0; i < l; i++) {

                if (getXMLTagValue((String) statusTag.elementAt(i), TAG_CMD).equals(statusOf)) {

                    String statusCode = getXMLTagValue((String) statusTag.elementAt(i), TAG_DATA);

                    if (String.valueOf(STATUS_CODE_OK).equals(statusCode)) {
                        //
                        // 200
                        //
                        return;
                    } else if (String.valueOf(STATUS_REFRESH_REQUIRED).equals(statusCode)) {
                        //
                        // 508
                        //
                        return;
                    } else if (String.valueOf(STATUS_CODE_AUTHENTICATION_ACCEPTED).equals(statusCode)) {
                        //
                        // 212
                        //
                        return;
                    } else if (String.valueOf(STATUS_CODE_FORBIDDEN).equals(statusCode)) {
                        //
                        // 403
                        //
                        throw new AuthenticationException(
                            "Sorry, you are not authorized to synchronize "
                            + sourceUri
                            + ". Go to the preferences panel and make sure your account settings are correct."
                        );
                    } else {
                        //
                        // Unhandled status code
                        //

                        throw new UpdateException(statusCode);
                    }
                }

        }  // next i

        String msgE = msg;
        StaticDataHelper.log(msgE);
        throw new SyncException(msgE);
    }

    /**
     * Checks response status for the synchronized databases and saves their
     * serverAlerts. <br>
     * If this is the first sync for the source, the status code might change
     * according to the value of the PARAM_FIRST_TIME_SYNC_MODE configuration
     * property.<br>
     * If firstTimeSyncMode is not set, the alert is left unchanged. If it is
     * set to a value, the specified value is used instead.
     *
     * @param msg the message to be checked
     *
     * @throws AuthenticationException in case the request has not been
     *                                 authenticated by the server
     * @throws SyncException in case of other errors
     **/
    private void checkServerAlerts(String msg)
    throws SyncException, AuthenticationException {

        Vector xmlMsg    = null;
        Vector itemMsg   = null;
        Vector targetTag = null;

        String alert     = null;
        String item      = null;
        String dataStore = null;

        xmlMsg = new Vector();
        xmlMsg.addElement(msg);

        serverAlerts = new Hashtable();

        Vector alertTag = getXMLTag(getXMLTag(getXMLTag(xmlMsg, TAG_SYNCML),TAG_SYNCBODY), TAG_ALERT);

        int l = alertTag.size();

        for (int i=0; i < l; i++) {
            itemMsg = new Vector();

            alert = getXMLTagValue((String) alertTag.elementAt(i), TAG_DATA);

            item = getXMLTagValue((String) alertTag.elementAt(i), TAG_ITEM);

            itemMsg.addElement(item);

            targetTag = getXMLTag(itemMsg, TAG_TARGET);

            int m = targetTag.size();

            for (int j=0; j < m; j++) {
                dataStore = getXMLTagValue((String) targetTag.elementAt(j), TAG_LOC_URI);
            }

            StaticDataHelper.log("The server alert code for " + dataStore + " is " + alert);

            serverAlerts.put(dataStore, alert);
        }
    }

    /**
     * Prepare inizialization message
     *
     **/
    private void prepareInizializationMessage()
    throws SyncException {

        String dbAlertsXML = null ;

        dbAlertsXML    = createAlerts(sourceUri)   ;

        //
        // authentication basic
        //
        login = new String(Base64.encode(login.getBytes()));

        clientInitXML  = messageFormat(clientInitXML, "{0}", sessionID    );
        clientInitXML  = messageFormat(clientInitXML, "{1}", serverUrl    );
        clientInitXML  = messageFormat(clientInitXML, "{2}", login        );
        clientInitXML  = messageFormat(clientInitXML, "{3}", MAX_MSG_SIZE );
        clientInitXML  = messageFormat(clientInitXML, "{4}", dbAlertsXML  );

    }

    /**
     * Process the modifications from the received from server.
     *
     * @param modifications the modification message
     *
     * @return true if a response message is required, false otherwise
     *
     * @throws SyncException
     */
    private boolean processModifications(String modifications)
    throws SyncException {
        Vector  xmlResponse    = new Vector();
        Vector  bodyTags       = new Vector();
        Vector  addTags        = new Vector();
        Vector  replaceTags    = new Vector();
        Vector  deleteTags     = new Vector();

        String  key            = null        ;
        String  content        = null        ;

        String  syncMsg        = null        ;

        Record  record         = null        ;

        boolean ret            = false       ;

        xmlResponse.addElement(modifications);

        bodyTags    = getXMLTag(getXMLTag(xmlResponse, TAG_SYNCML), TAG_SYNCBODY  );
        addTags     = getXMLTag(bodyTags,                           TAG_ADD       );
        replaceTags = getXMLTag(bodyTags,                           TAG_REPLACE   );
        deleteTags  = getXMLTag(bodyTags,                           TAG_DELETE    );

        if (modifications.indexOf("<" + TAG_SYNC + ">") != - 1) {
            ret = true;
        }

        for (int i=0, l = addTags.size(); i < l; i++) {

            key     = getXMLTagValue((String) addTags.elementAt(i), TAG_LOC_URI   );

            content = getXMLTagValue((String) addTags.elementAt(i), TAG_DATA      );

            StaticDataHelper sdh = new StaticDataHelper();
            if (sdh.isEncode() && content != null) {
                try {
                    content = new String (Base64.decode(content), "UTF-8");
                } catch (Exception e) {
                    throw new SyncException(e.getMessage());
                }
            }

            content = StringTools.unescapeXml(content);

            record = new Record(key, ' ', content);
            String bbKey = "";

            try {
                bbKey = dataStore.setRecord(record, false).getKey();
            } catch(DataAccessException e) {
                throw new SyncException(e.getMessage());
            }

            ret = true;

            mappings.put(bbKey, key);

        }

        for (int i=0, l = replaceTags.size(); i < l; i++) {

            key     = getXMLTagValue((String) replaceTags.elementAt(i), TAG_LOC_URI );

            content = getXMLTagValue((String) replaceTags.elementAt(i), TAG_DATA    );

            StaticDataHelper sdh = new StaticDataHelper();
            if (sdh.isEncode() && content != null) {
                try {
                    content = new String (Base64.decode(content), "UTF-8");
                } catch (Exception e) {
                    throw new SyncException(e.getMessage());
                }
            }

            content = StringTools.unescapeXml(content);

            record = new Record(key, ' ', content);

            try {
                dataStore.setRecord(record, true);
            } catch(DataAccessException e) {
                throw new SyncException(e.getMessage());
            }

            ret = true;

        }

        for (int i=0, l = deleteTags.size(); i < l; i++) {

            key = getXMLTagValue((String) deleteTags.elementAt(i), TAG_LOC_URI  );

            record = new Record();

            record.setKey(key);

            try {
                dataStore.deleteRecord(record);
            } catch(DataAccessException e) {
                throw new SyncException(e.getMessage());
            }

            ret = true;

        }

        return ret;
    }

    /**
     * Prepare the modification message. If includeClientModification is true, 
     * the message will include the client modifications; otherwise it will 
     * just include the status of for the commands sent by the server in the
     * previous message. It also returns true if this is the last message
     * containing client modifictions (as side effect, if 
     * includeClientModification is false, it returns always false).
     *
     * @param includeClientModifications true if the message shall include 
     *                                   client modifications
     * @param finalTag should <Final/> be included?
     *        
     * @return true if not all client modifications are going to be sent, 
     *         flase otherwise.
     **/
    private boolean prepareModificationMessage(boolean includeClientModifications)
    throws SyncException {

        Vector records = new Vector(DataStore.MAX_ITEM_NUMBER);
        boolean moreRecords = true;

        if (includeClientModifications) {

            String alert;
    
            int alertCode;
    
            //
            // Note we must sync only the sources that are aknoledged by the
            // server with a server Alert command.
            //
            StaticDataHelper.log("serverAlerts: " + serverAlerts);

            alert = (String)serverAlerts.get(sourceUri);
    
            if (alert != null) {
    
                alertCode = getSourceAlertCode(sourceUri);
    
                if (alertCode == ALERT_CODE_SLOW) {
                    //
                    // Slow Sync!
                    //
                    StaticDataHelper.log("Prepare slow sync for " + sourceUri);
                    moreRecords = filterRecordsForSlowSync(records);
                } else if (alertCode == ALERT_CODE_REFRESH_FROM_SERVER) {
                    //
                    // Refresh from server
                    //
                    StaticDataHelper.log("Prepare refresh for " + sourceUri);
                    records = new Vector(); // no items sent for refrsh
                } else if (alertCode == ALERT_CODE_ONE_WAY_FROM_SERVER) {
                    //
                    // One-way sync from server: no needs of client modifications
                    //
                    StaticDataHelper.log("Prepare one-way sync for " + sourceUri);
                    records = new Vector(); // no items sent for one-way
                } else {
                    //
                    // Fast Sync!
                    //
                    StaticDataHelper.log("Prepare fast sync for " + sourceUri);
                    //
                    // NOTE: filterRecordsForFastSync returns items in updated state
                    //
                    moreRecords = filterRecordsForFastSync(records);
                }
            } else {
    
                StaticDataHelper.log( "The server did not sent an Alert for "   +
                                    sourceUri                                +
                                    ". The source will not be synchronized."
                                    );
            }
    
            StaticDataHelper sdh = new StaticDataHelper();
    
            if (sdh.isEncode()) {
                records = encodeB64Records(records);
            }
        }

        modificationsXML    = messageFormat(modificationsXMLTemplate, "{0}", sessionID                  );
        modificationsXML    = messageFormat(modificationsXML, "{1}", serverUrl                          );
        modificationsXML    = messageFormat(modificationsXML, "{2}", MAX_MSG_SIZE                       );
        modificationsXML    = messageFormat(modificationsXML, "{3}", serverUrlStatus                    );
        modificationsXML    = messageFormat(modificationsXML, "{4}", 
                                  prepareSyncTag(records, sourceUri,includeClientModifications) );
        modificationsXML    = messageFormat(modificationsXML, "{5}", ((!moreRecords) ? "<Final/>" : "")     );

        return moreRecords;
    }

    /**
     * Prepare alert message
     *
     **/
    private void prepareAlertMessage() {

        alertXML    = messageFormat(alertXML, "{0}", sessionID       );
        alertXML    = messageFormat(alertXML, "{1}", serverUrl       );
        alertXML    = messageFormat(alertXML, "{2}", serverUrl       );
        alertXML    = messageFormat(alertXML, "{3}", serverUrlStatus );

    }

    /**
     * Prepare mapping message
     *
     **/
    private void prepareMappingMessage() {

        String targetRef    = null ;
        String sourceRef    = null ;

        StringBuffer mapTag = new StringBuffer();

        int i = 0;

        for (Enumeration e = mappings.keys(); e.hasMoreElements(); ) {

            if (i == 0) {
                mapTag.append("<Map>\n");

                mapTag.append("<CmdID>0</CmdID>\n");
                mapTag.append("<Target>\n");
                mapTag.append("<LocURI>" + sourceUri + "</LocURI>\n");
                mapTag.append("</Target>\n");
                mapTag.append("<Source>\n");
                mapTag.append("<LocURI>" + sourceUri + "</LocURI>\n");
                mapTag.append("</Source>\n");
            }

            sourceRef = (String) e.nextElement();
            targetRef = (String) mappings.get(sourceRef);

            mapTag.append("<MapItem>\n");
            mapTag.append("<Target>\n");
            mapTag.append("<LocURI>" + targetRef + "</LocURI>\n");
            mapTag.append("</Target>\n");
            mapTag.append("<Source>\n");
            mapTag.append("<LocURI>" + sourceRef + "</LocURI>\n");
            mapTag.append("</Source>\n");
            mapTag.append("</MapItem>\n");

            i++;
        }

        if (i > 0) {
            mapTag.append("</Map>\n");
        }
        mapTag.append("<Final/>\n");

        mappingXML    = messageFormat(mappingXML, "{0}", sessionID          );
        mappingXML    = messageFormat(mappingXML, "{1}", serverUrl          );
        mappingXML    = messageFormat(mappingXML, "{2}", serverUrl          );
        mappingXML    = messageFormat(mappingXML, "{3}", mapTag.toString()  );
    }

    /**
     * Load the resources from the classpath.
     *
     * @throws IOException
     **/
    private void loadResources() throws IOException {

        Class c = this.getClass();
/*
        alertXML = StaticDataHelper.read(
            c.getResourceAsStream(XML_ALERT)
        );
*/
        clientInitXML = StaticDataHelper.read(
            c.getResourceAsStream(XML_INITIALIZATION)
        );

        modificationsXMLTemplate = StaticDataHelper.read(
            c.getResourceAsStream(XML_MODIFICATIONS)
        );
        
        mappingXML = StaticDataHelper.read(
            c.getResourceAsStream(XML_MAPPING)
        );
    }

    /**
     * Contructs the alerts for the given databses.
     * @param sourceUri
     * @return the XML for the SyncML Alert commands
     **/
    private String createAlerts(String sourceUri) {

        StringBuffer sb = new StringBuffer();

        sb.append("<Alert>\n");
        sb.append("<CmdID>1</CmdID>\n");
        sb.append("<Data>");
        sb.append("200");
        sb.append("</Data>\n");
        sb.append("<Item>\n");
        sb.append("<Target><LocURI>");
        sb.append(sourceUri);
        sb.append("</LocURI></Target>\n");
        sb.append("<Source><LocURI>");
        sb.append(sourceUri);
        sb.append("</LocURI></Source>\n");
        sb.append("<Meta>\n");
        sb.append("<Anchor xmlns=\"syncml:metinf\">\n");
        sb.append("<Last>" + lastTimestamp + "</Last>\n");
        sb.append("<Next>" + nextTimestamp + "</Next>\n");
        sb.append("</Anchor>\n");
        sb.append("</Meta>\n");
        sb.append("</Item>\n");
        sb.append("</Alert>");
        sb.append("\n");

        return sb.toString();
    }

   /**
     * Make a String[] by tags find with search.
     *
     * @param xmlInput tags about search
     * @param tag to find
     * @return find tags
     **/
    private Vector getXMLTag(Vector xmlInput, String tag)
    throws SyncException {

        Vector xmlReturn    = null ;
        String xmlInputTag  = null ;
        String endTag       = null ;

        endTag   = "</" + tag + ">" ;

        xmlReturn = new Vector();

        try {

            for (int j=0, l = xmlInput.size(); j < l; j++) {

                xmlInputTag = (String) xmlInput.elementAt(j);

                //
                // tag without namespace
                // or tag with namespace
                //
                while (xmlInputTag.indexOf("<" + tag + ">") != -1 ||
                       xmlInputTag.indexOf("<" + tag + " ") != -1) {

                    xmlReturn.addElement(getXMLTagValue(xmlInputTag, tag));

                    xmlInputTag =
                        xmlInputTag.substring(
                                        xmlInputTag.
                                            indexOf(endTag) + endTag.length());

                }

            }

        } catch (Exception e) {
            throw new SyncException("Parsing XML error, TAG " +
                                    tag                       +
                                    " - : "                   +
                                    e.getMessage()            );
        }

        return xmlReturn;

    }

    /**
     * Make a String by value of <i>tag</i>.
     *
     * @param xml xml msg
     * @param tag tag to find + sourceType + 
     * @return tag value
     **/
    private String getXMLTagValue(String xml, String tag) {

        String startTag  = null;
        String endTag    = null;

        startTag = "<" + tag + ">";
        endTag   = "</" + tag + ">";

        return xml.substring(xml.indexOf(startTag) + startTag.length(), xml.indexOf(endTag));

    }

    /**
     * Retrieves the modified items since the last synchronization.
     *
     * @return the modified records since the last synchronization
     **/
    private boolean filterRecordsForFastSync(Vector records)
    throws SyncException {
        boolean moreRecords = false;

        if (moreRecordsNew) {
            try {
                moreRecords |= dataStore.getNextRecords(records, RECORD_STATE_NEW);
                moreRecordsNew = moreRecords;
                if (moreRecords) return moreRecords;
                dataStore.resetModificationCursor();
            } catch (DataAccessException e) {
                throw new SyncException(e.getMessage());
            }

        }
        
        if (moreRecordsUpdate) {
            try {
                moreRecords |= dataStore.getNextRecords(records, RECORD_STATE_UPDATED);
                moreRecordsUpdate = moreRecords;
                if (moreRecords) return moreRecords;
                dataStore.resetModificationCursor();
            } catch (DataAccessException e) {
                throw new SyncException(e.getMessage());
            }
        }
        
        if (moreRecordsDelete) {
            try {
                moreRecords |= dataStore.getNextRecords(records, RECORD_STATE_DELETED);
                moreRecordsDelete = moreRecords;
                if (moreRecords) return moreRecords;
                dataStore.resetModificationCursor();
            } catch (DataAccessException e) {
                throw new SyncException(e.getMessage());
            }
        }
        
        return moreRecords;

    }

    /**
     * Retrieves all items belonging to the given datastore, regardless their
     * modification status. It is used for slow sync.
     *
     * @return all items in the source
     **/
    private boolean filterRecordsForSlowSync(Vector records)
    throws SyncException {
        try {
            boolean moreElements = dataStore.getNextRecords(records);
    
            for(int i=0, l = records.size(); i < l; ++i) {
                ((Record) records.elementAt(i)).setState(RECORD_STATE_UPDATED);
            }
    
            return moreElements;
        } catch (Exception e) {
            throw new SyncException(e.getMessage());
        }
    }
    

    /**
     * return Sync tag about sourceUri
     *
     * @param records records to sync
     * @param sourceURI source uri
     * @return sync tag value
     **/
    private String prepareSyncTag(Vector records, String sourceURI, boolean sendingClientModifications)
    throws SyncException {

        StringBuffer addItems     = new StringBuffer() ;
        StringBuffer replaceItems = new StringBuffer() ;
        StringBuffer deleteItems  = new StringBuffer() ;
        StringBuffer add          = new StringBuffer() ;
        StringBuffer replace      = new StringBuffer() ;
        StringBuffer delete       = new StringBuffer() ;
        StringBuffer syncTag      = new StringBuffer() ;

        Record record = null;

        for(int i=0, l = records.size(); i < l; i++) {

            record = (Record) records.elementAt(i);

            try  {

                switch (record.getState()) {
                    case RECORD_STATE_DELETED:
                        deleteItems.append("<Item>\n")
                                .append("<Source><LocURI>")
                                .append(record.getKey())
                                .append("</LocURI></Source>\n")
                                .append("</Item>\n");
                        break;

                    case RECORD_STATE_UPDATED:

                        replaceItems.append("<Item>\n")
                                    .append("<Source><LocURI>")
                                    .append(record.getKey())
                                    .append("</LocURI></Source>\n")
                                    .append("<Data>")
                                    .append(StringTools.escapeXml(record.getUid()))
                                    .append("</Data>\n")
                                    .append("</Item>\n");

                        break;

                    case RECORD_STATE_NEW:

                        addItems.append("<Item>\n")
                                .append("<Source><LocURI>")
                                .append(record.getKey())
                                .append("</LocURI></Source>\n")
                                .append("<Data>")
                                .append(StringTools.escapeXml(record.getUid()))
                                .append("</Data>\n")
                                .append("</Item>\n");

                        break;


                } // end switch


            } catch (OutOfMemoryError oom) {
                throw new SyncException ("out-of-memory");
            }


        }  // next i

        //
        // NOTE: for JDK 1.1.8 compatibility we cannot use StringBuffer.append(StringBuffer)
        //

        if (addItems.length()>0) {
            add.append("<Add>\n")
               .append("<CmdID>4</CmdID>\n")
               .append("<Meta><Type xmlns=\"syncml:metinf\">").append(sourceType).append("</Type></Meta>\n")
               .append(addItems.toString())
               .append("\n</Add>\n");

       }

       if (replaceItems.length()>0) {
          replace.append("<Replace>\n")
                 .append("<CmdID>4</CmdID>\n")
                 .append("<Meta><Type xmlns=\"syncml:metinf\">").append(sourceType).append("</Type></Meta>\n")
                 .append(replaceItems.toString())
                 .append("\n</Replace>\n");

       }

       if (deleteItems.length()>0) {
          delete.append("<Delete>\n")
                 .append("<CmdID>4</CmdID>\n")
                 .append(deleteItems.toString())
                 .append("\n</Delete>\n");
       }

       syncTag.append("<Status>\n")
              .append("<CmdID>2</CmdID>\n")
              .append("<MsgRef>1</MsgRef><CmdRef>1</CmdRef><Cmd>Alert</Cmd>\n")
              .append("<TargetRef>")
              .append(sourceURI)
              .append("</TargetRef>\n")
              .append("<SourceRef>")
              .append(sourceURI)
              .append("</SourceRef>\n")
              .append("<Data>200</Data>\n")
              .append("<Item>\n")
              .append("<Data>" + "\n")
              .append("<Anchor xmlns=\"syncml:metinf\"><Next>")
              .append(nextTimestamp)
              .append("</Next></Anchor>\n")
              .append("</Data>\n")
              .append("</Item>\n")
              .append("</Status>\n");
              
      if (sendingClientModifications) {
          syncTag.append("<Sync>\n")
                 .append("<CmdID>3</CmdID>\n")
                 .append("<Target><LocURI>")
                 .append(sourceURI)
                 .append("</LocURI></Target>\n")
                 .append("<Source><LocURI>")
                 .append(sourceURI)
                 .append("</LocURI></Source>\n")
                 .append(add.toString())
                 .append(replace.toString())
                 .append(delete.toString())
                 .append("</Sync>");
      }

      add = null;
      replace = null;
      delete = null ;

       return syncTag.toString();
    }

    /**
     * Returns the server alert code for the given source.
     *
     * @param sourceURI the source
     *
     * @return the server alert code for the given source or -1 if it is not
     *         found/parsable
     */
    private int getSourceAlertCode(String sourceURI) {
        try {
            return Integer.parseInt((String)serverAlerts.get(sourceURI));
        } catch (Throwable t) {
            StaticDataHelper.log( "ERROR: unrecognized server alert code ("
                                + serverAlerts.get(sourceURI)
                                + ") for "
                                + sourceURI
                                );
        }

        return -1;
    }

    /**
     * Set variable in XML msg.
     *
     * @param msgXML msg XML
     * @param variable variable name
     * @param variableValue variable value
     * @return msg XML with setting variable value
     **/
    private String messageFormat(String msgXML, String variable, String variableValue) {

        String msgXMLBefore = null;
        String msgXMLAfter  = null;

        msgXMLBefore = msgXML.substring(0, msgXML.indexOf(variable));
        msgXMLAfter  = msgXML.substring(msgXML.indexOf(variable) + variable.length());

        return (msgXMLBefore + variableValue + msgXMLAfter);

    }

    private Vector encodeB64Records(Vector records) {

        String content = null;

        for (int i = 0, l = records.size(); records != null && i < l; i++) {

            content = ((Record) records.elementAt(i)).getUid();

            if (content != null && content.length() > 0) {
                ((Record) records.elementAt(i)).setUid(new String(Base64.encode(content.getBytes())));
            }

        }

        return records;
    }
}
